@@ -1,7 +1,7 @@
 /*
  * AIDE (Advanced Intrusion Detection Environment)
  *
- * Copyright (C) 1999-2002, 2004-2005, 2013, 2016, 2020 Rami Lehti,
+ * Copyright (C) 1999-2002, 2004-2005, 2013, 2016, 2020, 2022, 2023 Rami Lehti,
  *               Pablo Virolainen, Richard van den Berg, Hannes von Haugwitz
  *
  * This program is free software; you can redistribute it and/or
@@ -22,9 +22,11 @@
 #ifndef _DB_H_INCLUDED
 #define _DB_H_INCLUDED
 
-#include <stdio.h>
 #include <stdbool.h>
+#include <stdio.h>
+#include <time.h>
 #include "db_config.h"
+#include "util.h"
 
 typedef enum {
     DB_TYPE_IN,
@@ -48,19 +50,17 @@ void db_close();
 void free_db_line(db_line* dl);
 
 #define DB_OLD            (1<<0)
-#define DB_WRITE          (1<<1)
+#define DB_DISK           (1<<1)
 #define DB_NEW            (1<<2)
 #define NODE_ADDED        (1<<4)
 #define NODE_REMOVED      (1<<5)
 #define NODE_CHANGED      (1<<6)
 #define NODE_FREE         (1<<7)
-#define DB_DISK           (1<<8)
 
-#define NODE_TRAVERSE     (1<<9)
-#define NODE_CHECKED      (1<<10)
 #define NODE_MOVED_OUT    (1<<11)
 #define NODE_MOVED_IN     (1<<12)
 #define NODE_ALLOW_NEW    (1<<13)
 #define NODE_ALLOW_RM	  (1<<14)
+#define NODE_CHECK_INODE_CHILDS (1<<15)
 
 #endif
