@@ -1,7 +1,7 @@
 /*
  * AIDE (Advanced Intrusion Detection Environment)
  *
- * Copyright (C) 1999-2002, 2004-2006, 2010-2013, 2015-2016, 2019-2021
+ * Copyright (C) 1999-2002, 2004-2006, 2010-2013, 2015-2016, 2019-2022
  *               Rami Lehti, Pablo Virolainen, Richard van den Berg,
  *               Hannes von Haugwitz
  *
@@ -22,74 +22,27 @@
  
 #ifndef _DB_CONFIG_H_INCLUDED
 #define _DB_CONFIG_H_INCLUDED
+
 #include "config.h"
-#include "attributes.h"
-#include "report.h"
-#include "types.h"
-#include <unistd.h>
+#define PCRE2_CODE_UNIT_WIDTH 8
+#include <pcre2.h>
+#include <stdbool.h>
 #include <stdio.h>
-#include <pcre.h>
-
-#define E2O(n) (1<<n)
-
-#include "list.h"
-
-#ifdef WITH_POSIX_ACL /* POSIX acl works for Sun ACL, AIUI but anyway... */
-#include <sys/acl.h>
-#ifndef WITH_ACL
-# error "No ACL support ... but POSIX ACL support."
-#endif
-#endif
-
-typedef struct acl_type {
- char *acl_a; /* ACCESS */
- char *acl_d; /* DEFAULT, directories only */
-} acl_type;
-
-#ifdef WITH_XATTR /* Do generic user Xattrs. */
-#include <sys/xattr.h>
-#include <attr/attributes.h>
-#ifndef ENOATTR
-# define ENOATTR ENODATA
-#endif
-#endif
-
-typedef struct xattr_node 
-{
- char *key;
- byte *val;
- size_t vsz;
-} xattr_node;
-
-typedef struct xattrs_type
-{
-  size_t num;
-  size_t sz;
-  struct xattr_node *ents;
-} xattrs_type;
-
-#ifdef WITH_SELINUX
-#include <selinux/selinux.h>
-#ifndef ENOATTR
-# define ENOATTR ENODATA 
-#endif
-#endif
-
-#ifdef WITH_E2FSATTRS
-#include <e2p/e2p.h>
-#endif
-
-#ifdef WITH_CAPABILITIES
-#include <sys/capability.h>
-#endif
-
-#ifdef WITH_MHASH
-#include <mhash.h>
-#endif
-
+#include <sys/types.h>
 #ifdef WITH_ZLIB
 #include <zlib.h>
 #endif
+#include "attributes.h"
+#include "hashsum.h"
+#include "db_line.h"
+#include "list.h"
+#include "report.h"
+#include "rx_rule.h"
+#include "util.h"
+#include "url.h"
+
+
+#define E2O(n) (1<<n)
 
 #define RETOK 0
 #define RETFAIL -1
@@ -99,8 +52,6 @@ typedef struct xattrs_type
 #define DO_DIFF     (1<<2)
 #define DO_DRY_RUN  (1<<3)
 
-#include "url.h"
-
 /* TIMEBUFSIZE should be exactly ceil(sizeof(time_t)*8*ln(2)/ln(10))
  * Now it is ceil(sizeof(time_t)*2.5)
  * And of course we add one for end of string char
@@ -127,46 +78,6 @@ typedef struct xattrs_type
 /*    void* local; */  
 /*  }_db_config ; */
 
-
-#include "seltree.h"
-
-#include "hashsum.h"
-
-typedef struct db_line {
-  byte* hashsums[num_hashes];
-
-  acl_type* acl;
-  /* Something here.. */
-
-  mode_t perm;
-  mode_t perm_o; /* Permission for tree traverse */
-  long uid; /* uid_t */
-  long gid; /* gid_t */
-  time_t atime;
-  time_t ctime;
-  time_t mtime;
-  long inode; /* ino_t */
-  long nlink; /* nlink_t */
-
-  long long size; /* off_t */
-  long long bcount; /* blkcnt_t */
-  char* filename;
-  char* fullpath;
-  char* linkname;
-
-  char *cntx;
-
-  xattrs_type* xattrs;
-
-  unsigned long e2fsattrs;
-
-  char* capabilities;
-
-  /* Attributes .... */
-  DB_ATTR_TYPE attr;
-
-} db_line;
-
 typedef struct database {
     url_t* url;
 
@@ -210,6 +121,7 @@ typedef struct db_config {
   
   char* config_file;
   char* config_version;
+  bool config_check_warn_unrestricted_rules;
 
   int database_add_metadata;
   int report_detailed_init;
@@ -228,6 +140,7 @@ typedef struct db_config {
 
   list* report_urls;
   REPORT_LEVEL report_level;
+  REPORT_FORMAT report_format;
 
   /* defsyms is a list of symba*s */
   list* defsyms;
@@ -237,8 +150,9 @@ typedef struct db_config {
   /* What are we supposed to do */
   int action;
 
-  /* Should we catch errors from mmapping */
-  int catch_mmap;
+#ifdef WITH_PTHREAD
+  long num_workers;
+#endif
 
   time_t start_time;
   time_t end_time;
@@ -259,7 +173,8 @@ typedef struct db_config {
   int root_prefix_length;
 
   char* limit;
-  pcre* limit_crx;
+  pcre2_code* limit_crx;
+  pcre2_match_data* limit_md;
 
   struct seltree* tree;
 
