@@ -1,7 +1,7 @@
 /*
  * AIDE (Advanced Intrusion Detection Environment)
  *
- * Copyright (C) 1999-2006, 2010-2013, 2015-2017, 2019-2021 Rami Lehti,
+ * Copyright (C) 1999-2006, 2010-2013, 2015-2017, 2019-2023 Rami Lehti,
  *               Pablo Virolainen, Mike Markley, Richard van den Berg,
  *               Hannes von Haugwitz
  *
@@ -20,36 +20,37 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
-#include "aide.h"
+#include "config.h"
 
-#include <sys/types.h>
 #include <sys/stat.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <errno.h>
 #include <signal.h>
-#include <sys/types.h>
 #include <dirent.h>
 #include <time.h>
+#include <stdbool.h>
+#include <string.h>
 
 #if HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 
-#include "md.h"
+#include "attributes.h"
+#include "hashsum.h"
+#include "rx_rule.h"
+#include "url.h"
 #include "commandconf.h"
 #include "report.h"
 #include "db_config.h"
 #include "db_disk.h"
 #include "db.h"
-#include "do_md.h"
 #include "log.h"
+#include "seltree.h"
 #include "errorcodes.h"
 #include "gen_list.h"
 #include "getopt.h"
-#include "list.h"
 #include "util.h"
-#include "base64.h"
 /*for locale support*/
 #include "locale-aide.h"
 /*for locale support*/
@@ -61,10 +62,15 @@ char* after = NULL;
 #define MAXHOSTNAMELEN 256
 #endif
 
+#ifdef WITH_GCRYPT
+#include <gcrypt.h>
+#define NEED_LIBGCRYPT_VERSION "1.8.0"
+#endif
+
 static void usage(int exitvalue)
 {
   fprintf(stdout,
-	  _("Aide " AIDEVERSION" \n\n"
+	  _("AIDE %s \n\n"
 	    "Usage: aide [options] command\n\n"
 	    "Commands:\n"
 	    "  -i, --init\t\tInitialize the database\n"
@@ -74,16 +80,17 @@ static void usage(int exitvalue)
 	    "  -E, --compare\t\tCompare two databases\n\n"
 	    "Miscellaneous:\n"
 	    "  -D,\t\t\t--config-check\t\t\tTest the configuration file\n"
-	    "  -p file_type:path\t--path-check=file_type:path\tMatch file type and path against rule tree\n"
+	    "  -p FILE_TYPE:PATH\t--path-check=FILE_TYPE:PATH\tMatch file type and path against rule tree\n"
 	    "  -v,\t\t\t--version\t\t\tShow version of AIDE and compilation options\n"
 	    "  -h,\t\t\t--help\t\t\t\tShow this help message\n\n"
 	    "Options:\n"
-	    "  -c [cfgfile]\t--config=[cfgfile]\tGet config options from [cfgfile]\n"
-	    "  -l [REGEX]\t--limit=[REGEX]\t\tLimit command to entries matching [REGEX]\n"
+	    "  -c CFGFILE\t--config=CFGFILE\tGet config options from CFGFILE\n"
+	    "  -l REGEX\t--limit=REGEX\t\tLimit command to entries matching REGEX\n"
 	    "  -B \"OPTION\"\t--before=\"OPTION\"\tBefore configuration file is read define OPTION\n"
 	    "  -A \"OPTION\"\t--after=\"OPTION\"\tAfter configuration file is read define OPTION\n"
-	    "  -L [level]\t--log-level=[level]\tSet log message level to [level]\n"
-	    "\n")
+	    "  -L LEVEL\t--log-level=LEVEL\tSet log message level to LEVEL\n"
+	    "  -W WORKERS\t--workers=WORKERS\tNumber of simultaneous workers (threads) for file attribute processing (i.a. hashsum calculation)"
+	    "\n"), AIDEVERSION
 	  );
   
   exit(exitvalue);
@@ -93,7 +100,7 @@ static void sig_handler(int);
 
 static void init_sighandler()
 {
-  signal(SIGBUS,sig_handler);
+  log_msg(LOG_LEVEL_DEBUG, "initialize signal handler for SIGTERM, SIGUSR1 and SIGHUP");
   signal(SIGTERM,sig_handler);
   signal(SIGUSR1,sig_handler);
   signal(SIGHUP,sig_handler);
@@ -101,61 +108,66 @@ static void init_sighandler()
   return;
 }
 
+static void init_crypto_lib() {
+/* libmhash does not need to be initialized */
+#ifdef WITH_GCRYPT
+  if(!gcry_check_version(NEED_LIBGCRYPT_VERSION)) {
+      log_msg(LOG_LEVEL_ERROR, "libgcrypt is too old (need %s, have %s)", NEED_LIBGCRYPT_VERSION, gcry_check_version (NULL));
+      exit(VERSION_MISMATCH_ERROR);
+  }
+  gcry_control(GCRYCTL_DISABLE_SECMEM, 0);
+  gcry_control(GCRYCTL_INITIALIZATION_FINISHED, 0);
+#endif
+}
+
 static void sig_handler(int signum)
 {
-  switch(signum){
-  case SIGBUS  : {
-    if(conf->catch_mmap==1){
-      log_msg(LOG_LEVEL_NOTICE, "Caught SIGBUS while mmapping. File was truncated while aide was running?");
-      conf->catch_mmap=0;
-    } else {
-      log_msg(LOG_LEVEL_ERROR, "Caught SIGBUS. Exiting");
-      exit(EXIT_FAILURE);
+    char *str;
+    switch(signum){
+        case SIGHUP :
+          str = "Caught SIGHUP. Ignoring\n";
+          (void) !write(STDERR_FILENO ,str, strlen(str));
+          break;
+        case SIGTERM :
+           str = "Caught SIGTERM. Use SIGKILL to terminate\n";
+           (void) !write(STDERR_FILENO ,str, strlen(str));
+           break;
+        case SIGUSR1 :
+           str = "Caught SIGUSR1, toggle debug level\n";
+           (void) !write(STDERR_FILENO ,str, strlen(str));
+           toogle_log_level(LOG_LEVEL_DEBUG);
+           break;
     }
-    break;
-  }
-  case SIGHUP : {
-    log_msg(LOG_LEVEL_INFO, "Caught SIGHUP");
-    break;
-  }
-  case SIGTERM : {
-    log_msg(LOG_LEVEL_INFO, "Caught SIGTERM. Use SIGKILL to terminate");
-    break;
-  }
-  case SIGUSR1 : {
-    log_msg(LOG_LEVEL_INFO, "Caught SIGUSR1, toggle debug level: set log level to %s", get_log_level_name(toogle_log_level(LOG_LEVEL_DEBUG)));
-    break;
-  }
-  }
-  init_sighandler();
-
-  return;
 }
 
+#define EXTRA_ATTR(attribute) fprintf(stdout, "%s: %s\n", attributes[attribute].config_name, extra_attributes&ATTR(attribute)?"yes":"no");
+
 static void print_version(void)
 {
-  fprintf(stdout,
-	  "Aide " AIDEVERSION "\n\n"
-	  "Compiled with the following options:\n\n" AIDECOMPILEOPTIONS "\n");
-
+  fprintf(stdout, "AIDE %s\n\n", AIDEVERSION );
+  fprintf(stdout, "Compile-time options:\n%s\n", AIDECOMPILEOPTIONS);
   fprintf(stdout, "Default config values:\n");
   fprintf(stdout, "config file: %s\n", conf->config_file?conf->config_file:"<none>");
-  fprintf(stdout, "database_in: %s\n",
 #ifdef DEFAULT_DB
-          DEFAULT_DB
+  fprintf(stdout, "database_in: %s\n", DEFAULT_DB);
 #else
-          "<none>"
+  fprintf(stdout, "database_in: <none>\n");
 #endif
-          ),
-  fprintf(stdout, "database_out: %s\n",
 #ifdef DEFAULT_DB_OUT
-          DEFAULT_DB_OUT
+  fprintf(stdout, "database_out: %s\n", DEFAULT_DB_OUT),
 #else
-          "<none>"
+  fprintf(stdout, "database_out: <none>\n"),
 #endif
-          ),
 
-  fprintf(stdout, "\nAvailable hashsum groups:\n");
+  fprintf(stdout, "\nAvailable compiled-in attributes:\n");
+  DB_ATTR_TYPE extra_attributes = get_groupval("X");
+  EXTRA_ATTR(attr_acl)
+  EXTRA_ATTR(attr_xattrs)
+  EXTRA_ATTR(attr_selinux)
+  EXTRA_ATTR(attr_e2fsattrs)
+  EXTRA_ATTR(attr_capabilities)
+
+  fprintf(stdout, "\nAvailable hashsum attributes:\n");
   DB_ATTR_TYPE available_hashsums = get_hashes(false);
   for (int i = 0; i < num_hashes; ++i) {
       fprintf(stdout, "%s: %s\n", attributes[hashsums[i].attribute].config_name, ATTR(hashsums[i].attribute)&available_hashsums?"yes":"no");
@@ -175,14 +187,13 @@ static void print_version(void)
 static char *append_line_to_config(char *config, char *line) {
     size_t line_length = strlen(line);
     if (config == NULL) {
-        config = checked_malloc((line_length+2)*sizeof(char));
-        strcpy(config,line);
-        strcat(config,"\n");
+        int len = (line_length + 2)*sizeof(char);
+        config = checked_malloc(len);
+        snprintf(config, len, "%s\n", line);
     } else {
-        char *tmp = checked_malloc(sizeof(char) *(strlen(config)+line_length+2));
-        strcpy(tmp,config);
-        strcat(tmp,line);
-        strcat(tmp,"\n");
+        int len = (strlen(config) + line_length + 2) * sizeof(char);
+        char *tmp = checked_malloc(len);
+        snprintf(tmp, len, "%s%s\n", config, line);
         free(config);
         config=tmp;
     }
@@ -200,16 +211,13 @@ static char *append_line_to_config(char *config, char *line) {
                 log_msg(LOG_LEVEL_INFO,"(%s): %s command", longopt, desc); \
             } else { \
                 INVALID_ARGUMENT(longopt, %s, "cannot have multiple commands on a single commandline") \
-                exit(INVALID_ARGUMENT_ERROR); \
             } \
             break; \
         }
 
 static void read_param(int argc,char**argv)
 {
-  int option = -1;
   int i=0;
-  
 
   static struct option options[] =
   {
@@ -228,12 +236,13 @@ static void read_param(int argc,char**argv)
     { "path-check", required_argument, NULL, 'p'},
     { "limit", required_argument, NULL, 'l'},
     { "log-level", required_argument, NULL, 'L'},
+    { "workers", required_argument, NULL, 'W'},
     { "compare", no_argument, NULL, 'E'},
     { NULL,0,NULL,0 }
   };
 
   while(1){
-    option = getopt_long(argc, argv, "hL:V::vc:l:p:B:A:riCuDEn", options, &i);
+    int option = getopt_long(argc, argv, "hL:V::vc:l:p:B:A:W:riCuDEn", options, &i);
     if(option==-1)
       break;
     switch(option)
@@ -248,7 +257,6 @@ static void read_param(int argc,char**argv)
       }
       case 'V':{
         INVALID_ARGUMENT("--verbose", %s, "option no longer supported, use 'log_level' and 'report_level' options instead (see man aide.conf for details)")
-	break;
       }
       case 'c':{
 	  conf->config_file=optarg;
@@ -266,13 +274,32 @@ static void read_param(int argc,char**argv)
 	break;
       }
       case 'l': {
-                const char* pcre_error;
-                int pcre_erroffset;
-                conf->limit=checked_malloc(strlen(optarg)+1);
-                strcpy(conf->limit,optarg);
-                if((conf->limit_crx=pcre_compile(conf->limit, PCRE_ANCHORED, &pcre_error, &pcre_erroffset, NULL)) == NULL) {
-                    INVALID_ARGUMENT("--limit", error in regular expression '%s' at %i: %s, conf->limit, pcre_erroffset, pcre_error)
+                int pcre2_errorcode;
+                PCRE2_SIZE pcre2_erroffset;
+                int len = (strlen(optarg)+1) * sizeof(char);
+                conf->limit=checked_malloc(len);
+                strncpy(conf->limit, optarg, len);
+                if((conf->limit_crx=pcre2_compile((PCRE2_SPTR) conf->limit, PCRE2_ZERO_TERMINATED, PCRE2_UTF|PCRE2_ANCHORED, &pcre2_errorcode, &pcre2_erroffset, NULL)) == NULL) {
+                    PCRE2_UCHAR pcre2_error[128];
+                    pcre2_get_error_message(pcre2_errorcode, pcre2_error, 128);
+                    INVALID_ARGUMENT("--limit", error in regular expression '%s' at %zu: %s, conf->limit, pcre2_erroffset, pcre2_error)
+
                 }
+                conf->limit_md = pcre2_match_data_create_from_pattern(conf->limit_crx, NULL);
+                if (conf->limit_md == NULL) {
+                    log_msg(LOG_LEVEL_ERROR, "pcre2_match_data_create_from_pattern: failed to allocate memory");
+                    exit(MEMORY_ALLOCATION_FAILURE);
+                }
+
+                int pcre2_jit = pcre2_jit_compile(conf->limit_crx, PCRE2_JIT_PARTIAL_SOFT);
+                if (pcre2_jit < 0) {
+                    PCRE2_UCHAR pcre2_error[128];
+                    pcre2_get_error_message(pcre2_jit, pcre2_error, 128);
+                    log_msg(LOG_LEVEL_NOTICE, "JIT compilation for limit '%s' failed: %s (fall back to interpreted matching)", conf->limit, pcre2_error);
+                } else {
+                    log_msg(LOG_LEVEL_DEBUG, "JIT compilation for limit '%s' successful", conf->limit);
+                }
+
                 log_msg(LOG_LEVEL_INFO,_("(--limit): set limit to '%s'"), conf->limit);
             break;
       }
@@ -286,6 +313,19 @@ static void read_param(int argc,char**argv)
             }
            break;
                }
+      case 'W':{
+#ifdef WITH_PTHREAD
+           long num_workers = do_num_workers(optarg);
+           if (num_workers < 0) {
+               INVALID_ARGUMENT("--workers", invalid number of workers '%s', optarg)
+           }
+           conf->num_workers = num_workers;
+           log_msg(LOG_LEVEL_INFO,"(--workers): set number of workers to %ld (argument value: '%s')", conf->num_workers, optarg);
+#else
+           INVALID_ARGUMENT("--workers", %s, "pthread support not compiled in, recompile AIDE with '--with-pthread'")
+#endif
+           break;
+      }
       case 'p':{
             if(conf->action==0){
                 conf->action=DO_DRY_RUN;
@@ -310,13 +350,11 @@ static void read_param(int argc,char**argv)
 
             } else {
                 INVALID_ARGUMENT("--path-check", %s, "cannot have multiple commands on a single commandline")
-                exit(INVALID_ARGUMENT_ERROR);
             }
             break;
       }
       case 'r': {
        INVALID_ARGUMENT("--report", %s, "option no longer supported, use 'report_url' config option instead (see man aide.conf for detail)")
-       break;
       }
       ACTION_CASE("--init", 'i', DO_INIT, "database init")
       ACTION_CASE("--dry-init", 'n', DO_INIT|DO_DRY_RUN, "dry init")
@@ -344,7 +382,7 @@ static void setdefaults_before_config()
 
   /* Setting some defaults */
 
-  log_msg(LOG_LEVEL_INFO, "initialise rule tree");
+  log_msg(LOG_LEVEL_INFO, "initialize rule tree");
   conf->tree=init_tree();
   conf->database_add_metadata=1;
   conf->report_detailed_init=0;
@@ -363,7 +401,8 @@ static void setdefaults_before_config()
   conf->check_file_type = FT_REG;
 
   conf->report_urls=NULL;
-  conf->report_level=REPORT_LEVEL_CHANGED_ATTRIBUTES;
+  conf->report_level=default_report_options.level;
+  conf->report_format=default_report_options.format;
 
   conf->config_file=
 #ifdef CONFIG_FILE
@@ -373,6 +412,7 @@ static void setdefaults_before_config()
 #endif
       ;
   conf->config_version=NULL;
+  conf->config_check_warn_unrestricted_rules = false;
   
 #ifdef WITH_ACL
   conf->no_acl_on_symlinks=0; /* zero means don't do ACLs on symlinks */
@@ -433,7 +473,10 @@ static void setdefaults_before_config()
 #endif
 
   conf->action=0;
-  conf->catch_mmap=0;
+
+#ifdef WITH_PTHREAD
+  conf->num_workers = -1;
+#endif
 
   conf->warn_dead_symlinks=0;
 
@@ -449,9 +492,9 @@ static void setdefaults_before_config()
 
   conf->groupsyms=NULL;
 
-  conf->start_time=time(&(conf->start_time));
+  conf->start_time=time(NULL);
 
-  log_msg(LOG_LEVEL_INFO, "define default group definitions");
+  log_msg(LOG_LEVEL_INFO, "define default attribute definitions");
 
   for (ATTRIBUTE i = 0 ; i < num_attrs ; ++i) {
       if (attributes[i].config_name) {
@@ -478,13 +521,24 @@ static void setdefaults_before_config()
 
   DB_ATTR_TYPE common_attrs = ATTR(attr_perm)|ATTR(attr_ftype)|ATTR(attr_inode)|ATTR(attr_linkcount)|ATTR(attr_uid)|ATTR(attr_gid);
 
-  do_groupdef("R",common_attrs|ATTR(attr_size)|ATTR(attr_linkname)|ATTR(attr_mtime)|ATTR(attr_ctime)
-#if defined(WITH_MHASH) || defined(WITH_GCRYPT)
-          |ATTR(attr_md5)
+  DB_ATTR_TYPE GROUP_R_HASHES=0LLU;
+#ifdef WITH_MHASH
+  GROUP_R_HASHES=ATTR(attr_md5);
 #endif
-          |X);
+#ifdef WITH_GCRYPT
+  if (gcry_fips_mode_active()) {
+    char* str;
+    log_msg(LOG_LEVEL_NOTICE, "libgcrypt is running in FIPS mode, the following hash(es) are not available: %s", str = diff_attributes(0, ATTR(attr_md5)));
+    free(str);
+  } else {
+    GROUP_R_HASHES = ATTR(attr_md5);
+  }
+#endif
+
+  log_msg(LOG_LEVEL_INFO, "define default groups definitions");
+  do_groupdef("R",common_attrs|ATTR(attr_size)|ATTR(attr_linkname)|ATTR(attr_mtime)|ATTR(attr_ctime)|GROUP_R_HASHES|X);
   do_groupdef("L",common_attrs|ATTR(attr_linkname)|X);
-  do_groupdef(">",common_attrs|ATTR(attr_sizeg)|ATTR(attr_linkname)|X);
+  do_groupdef(">",common_attrs|ATTR(attr_size)|ATTR(attr_growing)|ATTR(attr_linkname)|X);
   do_groupdef("H",get_hashes(false));
   do_groupdef("X",X);
   do_groupdef("E",0);
@@ -518,24 +572,34 @@ static void setdefaults_after_config()
     conf->action=DO_COMPARE;
   }
 
+#ifdef WITH_PTHREAD
+  if(conf->num_workers < 0) {
+      conf->num_workers = 1;
+      log_msg(LOG_LEVEL_CONFIG, "(default): set 'num_workers' option to %lu", conf->num_workers);
+  }
+#endif
+
   if (is_log_level_unset()) {
           set_log_level(LOG_LEVEL_WARNING);
   };
 }
 
-
-
 int main(int argc,char**argv)
 {
   int errorno=0;
 
-#ifdef USE_LOCALE
+#ifdef WITH_PTHREAD
+  log_init();
+#endif
+
+#ifdef WITH_LOCALE
   setlocale(LC_ALL,"");
   bindtextdomain(PACKAGE,LOCALEDIR);
   textdomain(PACKAGE);
 #endif
   umask(0177);
   init_sighandler();
+  init_crypto_lib();
 
   setdefaults_before_config();
 
@@ -568,14 +632,29 @@ int main(int argc,char**argv)
   log_msg(LOG_LEVEL_RULE, "rule tree:");
   log_tree(LOG_LEVEL_RULE, conf->tree, 0);
 
+  if (conf->action&DO_INIT
+          && (conf->tree)->childs == NULL
+          && (conf->tree)->equ_rx_lst == NULL
+          && (conf->tree)->sel_rx_lst == NULL
+          && (conf->tree)->neg_rx_lst == NULL
+     ) {
+      log_msg(LOG_LEVEL_WARNING, "rule tree is empty, no files will be added to the database");
+  }
+
   if (conf->check_path) {
       rx_rule* rule = NULL;
-      int match = check_rxtree(conf->check_path, conf->tree, &rule, conf->check_file_type, true);
-      if (match < 0) {
-        fprintf(stdout, "[ ] %c '%s': outside of limit '%s'\n", get_restriction_char(conf->check_file_type), conf->check_path, conf->limit);
-        exit(2);
-      } else {
-        exit(match?0:1);
+      match_result match = check_rxtree(conf->check_path, conf->tree, &rule, conf->check_file_type, "disk (path-check)");
+      print_match(conf->check_path, rule, match, conf->check_file_type);
+      switch (match) {
+          case RESULT_PARTIAL_LIMIT_MATCH:
+          case RESULT_NO_LIMIT_MATCH:
+              exit(2);
+          case RESULT_EQUAL_MATCH:
+          case RESULT_SELECTIVE_MATCH:
+              exit(0);
+          case RESULT_NO_MATCH:
+          case RESULT_PARTIAL_MATCH:
+              exit(1);
       }
   }
 
@@ -607,12 +686,14 @@ int main(int argc,char**argv)
     exit(INVALID_ARGUMENT_ERROR);
   }
 
+  /* ensure size attribute is added to db_out_attrs if sizeg or growing attribute is set */
+  if (conf->db_out_attrs & ATTR(attr_sizeg) || conf->db_out_attrs & ATTR(attr_growing)) {
+        conf->db_out_attrs |=ATTR(attr_size);
+  }
+
   if (conf->action&DO_INIT && conf->action&DO_DRY_RUN) {
-      if(db_disk_init()==RETFAIL) {
-          exit(IO_ERROR);
-      }
-      log_msg(LOG_LEVEL_INFO, "populate tree (dry-run)");
-      populate_tree(conf->tree, true);
+      log_msg(LOG_LEVEL_INFO, "scan file system (dry-run)");
+      db_scan_disk(true);
       exit (0);
   }
 
@@ -646,10 +727,6 @@ int main(int argc,char**argv)
 	exit(IO_ERROR);
       }
     }
-    if((conf->action&DO_INIT)||(conf->action&DO_COMPARE)){
-      if(db_disk_init()==RETFAIL)
-	exit(IO_ERROR);
-    }
     if((conf->action&DO_COMPARE)||(conf->action&DO_DIFF)){
       if(db_init(&(conf->database_in), true, false)==RETFAIL)
 	exit(IO_ERROR);
@@ -658,9 +735,23 @@ int main(int argc,char**argv)
       if(db_init(&(conf->database_new), true, false)==RETFAIL)
 	exit(IO_ERROR);
     }
-      
+
+#ifdef WITH_PTHREAD
+    if((conf->action&DO_INIT || conf->action&DO_COMPARE) && conf->num_workers){
+      if(db_disk_start_threads()==RETFAIL)
+          exit(THREAD_ERROR);
+    }
+#endif
+
     log_msg(LOG_LEVEL_INFO, "populate tree");
-    populate_tree(conf->tree, false);
+    populate_tree(conf->tree);
+
+#ifdef WITH_PTHREAD
+    if((conf->action&DO_INIT || conf->action&DO_COMPARE) && conf->num_workers){
+      if(db_disk_finish_threads() == RETFAIL)
+          exit(THREAD_ERROR);
+    }
+#endif
 
     if(conf->action&DO_INIT) {
         log_msg(LOG_LEVEL_INFO, "write new entries to database: %s:%s", get_url_type_string((conf->database_out.url)->type), (conf->database_out.url)->value);
@@ -669,6 +760,8 @@ int main(int argc,char**argv)
 
     db_close();
 
+    conf->end_time=time(NULL);
+
     log_msg(LOG_LEVEL_INFO, "generate reports");
 
     int exitcode = gen_report(conf->tree);
