@@ -1,7 +1,7 @@
 /*
  * AIDE (Advanced Intrusion Detection Environment)
  *
- * Copyright (C) 1999-2006, 2010-2011, 2013, 2019-2021 Rami Lehti,
+ * Copyright (C) 1999-2006, 2010-2011, 2013, 2019-2023 Rami Lehti,
  *               Pablo Virolainen, Richard van den Berg, Hannes von Haugwitz
  *
  * This program is free software; you can redistribute it and/or
@@ -20,13 +20,19 @@
  */
  
 #include "aide.h"
+#include <stdbool.h>
 #include <stdio.h>
+#include <string.h>
+#include <time.h>
+#include "attributes.h"
+#include "config.h"
+#include "hashsum.h"
+#include "url.h"
 #include <stdlib.h>
-#include <errno.h>
 #include "db.h"
+#include "db_line.h"
 #include "db_lex.h"
 #include "db_file.h"
-#include "db_disk.h"
 #include "md.h"
 
 #ifdef WITH_CURL
@@ -37,15 +43,8 @@
 #include "log.h"
 #include "be.h"
 
-#ifdef WITH_MHASH
-#include <mhash.h>
-#endif
-
 #include "base64.h"
 #include "util.h"
-/*for locale support*/
-#include "locale-aide.h"
-/*for locale support*/
 
 db_line* db_char2line(char**, database*);
 
@@ -54,7 +53,7 @@ static long readoct(char* s, database* db, char* field_name){
   char* e;
   i=strtol(s,&e,8);
   if (e[0]!='\0') {
-      LOG_DB_FORMAT_LINE(LOG_LEVEL_WARNING, could not read '%s' from database: strtol (base: 8) failed for '%s', field_name, s)
+      LOG_DB_FORMAT_LINE(LOG_LEVEL_WARNING, "could not read '%s' from database: strtol (base: 8) failed for '%s'", field_name, s)
   }
   return i;
 }
@@ -64,7 +63,7 @@ static long readlong(char* s, database* db, char* field_name){
   char* e;
   i=strtol(s,&e,10);
   if (e[0]!='\0') {
-      LOG_DB_FORMAT_LINE(LOG_LEVEL_WARNING, could not read '%s' from database: strtol failed for '%s', field_name, s)
+      LOG_DB_FORMAT_LINE(LOG_LEVEL_WARNING, "could not read '%s' from database: strtol failed for '%s'", field_name, s)
   }
   return i;
 }
@@ -74,7 +73,7 @@ static long long readlonglong(char* s, database* db, char* field_name){
   char* e;
   i=strtoll(s,&e,10);
   if (e[0]!='\0') {
-      LOG_DB_FORMAT_LINE(LOG_LEVEL_WARNING, could not read '%s' from database: strtoll failed for '%s', field_name, s)
+      LOG_DB_FORMAT_LINE(LOG_LEVEL_WARNING, "could not read '%s' from database: strtoll failed for '%s'", field_name, s)
   }
   return i;
 }
@@ -93,12 +92,7 @@ static struct md_container *init_db_attrs(url_t *u) {
             case url_ftp: {
                 mdc = checked_malloc(sizeof(struct md_container)); /* freed in close_db_attrs */
                 mdc->todo_attr = conf->db_attrs;
-
-                int length = snprintf(NULL, 0, "%s:%s", get_url_type_string(u->type), u->value) + 1;
-                char *str = checked_malloc(length * sizeof(char));
-                snprintf(str, length, "%s:%s", get_url_type_string(u->type), u->value);
-                init_md(mdc, str);
-                free(str);
+                init_md(mdc, u->value);
 
                 break;
             }
@@ -115,12 +109,13 @@ static struct md_container *init_db_attrs(url_t *u) {
 static db_line *close_db_attrs (database *db) {
     db_line *line = NULL;
     if (db->mdc != NULL) {
-        close_md(db->mdc);
+        md_hashsums hs;
+        close_md(db->mdc, &hs, (db->url)->value);
         line = checked_malloc(sizeof(struct db_line));
         line->filename = (db->url)->value;
         line->perm = 0;
         line->attr = conf->db_attrs;
-        md2line(db->mdc, line);
+        hashsums2line(&hs, line);
         free(db->mdc);
     }
     return line;
@@ -232,8 +227,12 @@ db_line* db_char2line(char** ss, database* db){
   line->filename=NULL;
   line->fullpath=NULL;
   line->linkname=NULL;
+#ifdef WITH_POSIX_ACL
   line->acl=NULL;
+#endif
+#ifdef WITH_XATTR
   line->xattrs=NULL;
+#endif
   line->e2fsattrs=0;
   line->cntx=NULL;
   line->capabilities=NULL;
@@ -337,6 +336,7 @@ db_line* db_char2line(char** ss, database* db){
       break;
     }
       case attr_xattrs : {
+#ifdef WITH_XATTR
         size_t num = 0;
         char *tval = NULL;
         
@@ -364,6 +364,7 @@ db_line* db_char2line(char** ss, database* db){
             ++num;
           }
         }
+#endif
         break;
       }
 
@@ -409,6 +410,8 @@ db_line* db_char2line(char** ss, database* db){
     case attr_ftype :
     case attr_checkinode :
     case attr_allhashsums :
+    case attr_growing :
+    case attr_compressed :
     case attr_allownewfile :
     case attr_allowrmfile : {
       /*  no db field */
@@ -443,7 +446,7 @@ time_t base64totime_t(char* s, database* db, const char* field_name){
     time_t t = strtol((char *)b,&endp,10);
     
     if (endp[0]!='\0') {
-      LOG_DB_FORMAT_LINE(LOG_LEVEL_WARNING, could not read '%s' from database: strtoll failed for '%s' (base64 encoded value: '%s'), field_name, b, s)
+      LOG_DB_FORMAT_LINE(LOG_LEVEL_WARNING, "could not read '%s' from database: strtoll failed for '%s' (base64 encoded value: '%s')", field_name, b, s)
       free(b);
       return 0;
     }
@@ -542,17 +545,19 @@ void free_db_line(db_line* dl)
   checked_free(dl->fullpath);
   checked_free(dl->linkname);
   
+#ifdef WITH_ACL
   if (dl->acl)
   {
-#ifdef WITH_ACL
     free(dl->acl->acl_a);
     free(dl->acl->acl_d);
-#endif
   }
   checked_free(dl->acl);
+#endif
   
+#ifdef WITH_XATTR
   if (dl->xattrs)
     free(dl->xattrs->ents);
   checked_free(dl->xattrs);
   checked_free(dl->cntx);
+#endif
 }
