@@ -46,6 +46,10 @@
 #include <pthread.h>
 #endif
 
+#ifdef WITH_PTHREAD
+pthread_mutex_t seltree_mutex = PTHREAD_MUTEX_INITIALIZER;
+#endif
+
 static int get_file_status(char *filename, struct stat *fs) {
     int sres = 0;
     sres = lstat(filename,fs);
@@ -121,7 +125,13 @@ void scan_dir(char *root_path, bool dry_run) {
 
     log_msg(LOG_LEVEL_DEBUG,"scan_dir: process root directory '%s' (fullpath: '%s')", &root_path[conf->root_prefix_length], root_path);
     if (!get_file_status(root_path, &fs)) {
+#ifdef WITH_PTHREAD
+        pthread_mutex_lock(&seltree_mutex);
+#endif
         match_result match = check_rxtree (&root_path[conf->root_prefix_length], conf->tree, &rule, get_restriction_from_perm(fs.st_mode), "disk");
+#ifdef WITH_PTHREAD
+        pthread_mutex_unlock(&seltree_mutex);
+#endif
         if (dry_run) {
             print_match(&root_path[conf->root_prefix_length], rule, match, get_restriction_from_perm(fs.st_mode));
         }
@@ -152,7 +162,13 @@ void scan_dir(char *root_path, bool dry_run) {
                     if (!get_file_status(entry_full_path, &fs)) {
                         rule = NULL;
                         node = NULL;
+#ifdef WITH_PTHREAD
+                        pthread_mutex_lock(&seltree_mutex);
+#endif
                         match_result match = check_rxtree (&entry_full_path[conf->root_prefix_length], conf->tree, &rule, get_restriction_from_perm(fs.st_mode), "disk");
+#ifdef WITH_PTHREAD
+                        pthread_mutex_unlock(&seltree_mutex);
+#endif
                         switch (match) {
                             case RESULT_SELECTIVE_MATCH:
                                 if (S_ISDIR(fs.st_mode)) {
@@ -174,7 +190,13 @@ void scan_dir(char *root_path, bool dry_run) {
                                 }
                                 break;
                             case RESULT_NO_MATCH:
+#ifdef WITH_PTHREAD
+                                pthread_mutex_lock(&seltree_mutex);
+#endif
                                 node = get_seltree_node(conf->tree, &entry_full_path[conf->root_prefix_length]);
+#ifdef WITH_PTHREAD
+                                pthread_mutex_unlock(&seltree_mutex);
+#endif
                                 if(S_ISDIR(fs.st_mode) && node) {
                                     log_msg(log_level, "scan_dir: add child directory '%s' to scan stack (reason: existing tree node '%s' (%p))", &entry_full_path[conf->root_prefix_length], node->path, node);
                                     free_entry_full_path = false;
@@ -221,7 +243,9 @@ static void * add2tree( __attribute__((unused)) void *arg) {
     database_entry *data;
     while ((data = queue_ts_dequeue_wait(queue_database_entries, whoami)) != NULL) {
         log_msg(LOG_LEVEL_THREAD, "%10s: got line '%s'", whoami, (data->line)->filename);
+        pthread_mutex_lock(&seltree_mutex);
         add_file_to_tree(conf->tree, data->line, DB_NEW|DB_DISK, NULL, &data->fs);
+        pthread_mutex_unlock(&seltree_mutex);
         free(data);
     }
     queue_ts_free(queue_database_entries);
