@@ -1,7 +1,7 @@
 /*
  * AIDE (Advanced Intrusion Detection Environment)
  *
- * Copyright (C) 1999-2006, 2009-2011, 2015-2016, 2019-2021 Rami Lehti,
+ * Copyright (C) 1999-2006, 2009-2011, 2015-2016, 2019-2022 Rami Lehti,
  *               Pablo Virolainen, Richard van den Berg, Hannes von Haugwitz
  *
  * This program is free software; you can redistribute it and/or
@@ -19,17 +19,30 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
+#include <stdbool.h>
 #include <stdlib.h>
 #include <string.h>
-
-#include "seltree.h"
+#include "attributes.h"
+#include "list.h"
 #include "log.h"
+#include "rx_rule.h"
+#include "seltree.h"
+#include "seltree_struct.h"
 #include "util.h"
+#include "errorcodes.h"
 
-#define PARTIAL_RULE_MATCH       (-1)
-#define NO_RULE_MATCH            (0)
-#define RESTRICTED_RULE_MATCH    (1)
-#define RULE_MATCH               (2)
+#define NO_RULE_MATCH               0
+#define NEGATIVE_RULE_MATCH     (1<<0)
+#define SELECtIVE_RULE_MATCH    (1<<1)
+#define EQUAL_RULE_MATCH        (1<<2)
+#define RESTRICTED_RULE_MATCH   (1<<3)
+#define PARTIAL_RULE_MATCH      (1<<4)
+#define RULE_MATCH              (EQUAL_RULE_MATCH|SELECtIVE_RULE_MATCH)
+
+#define DEEP_EQUAL_MATCH        (1<<5)
+#define DEEP_SELECTIVE_MATCH    (1<<6)
+#define RECURSED_CALL           (1<<7)
+#define TOP_LEVEL_CALL          (1<<8)
 
 void log_tree(LOG_LEVEL log_level, seltree* tree, int depth) {
 
@@ -42,19 +55,19 @@ void log_tree(LOG_LEVEL log_level, seltree* tree, int depth) {
 
     for(r=tree->equ_rx_lst;r!=NULL;r=r->next) {
         rxc=r->data;
-        log_msg(log_level, "%-*s  '=%s %s %s' (%s:%d: '%s')", depth+2, "\u2502", rxc->rx, rs_str = get_restriction_string(rxc->restriction), attr_str = diff_attributes(0, rxc->attr), rxc->config_filename, rxc->config_linenumber, rxc->config_line);
+        log_msg(log_level, "%-*s  '=%s %s %s' (%s:%d: '%s%s%s')", depth+2, "\u2502", rxc->rx, rs_str = get_restriction_string(rxc->restriction), attr_str = diff_attributes(0, rxc->attr), rxc->config_filename, rxc->config_linenumber, rxc->config_line, rxc->prefix?"', prefix: '":"", rxc->prefix?rxc->prefix:"");
         free(rs_str);
         free(attr_str);
     }
     for(r=tree->sel_rx_lst;r!=NULL;r=r->next) {
         rxc=r->data;
-        log_msg(log_level, "%-*s  '%s %s %s' (%s:%d: '%s')", depth+2, "\u2502", rxc->rx, rs_str = get_restriction_string(rxc->restriction), attr_str = diff_attributes(0, rxc->attr), rxc->config_filename, rxc->config_linenumber, rxc->config_line);
+        log_msg(log_level, "%-*s  '%s %s %s' (%s:%d: '%s%s%s')", depth+2, "\u2502", rxc->rx, rs_str = get_restriction_string(rxc->restriction), attr_str = diff_attributes(0, rxc->attr), rxc->config_filename, rxc->config_linenumber, rxc->config_line, rxc->prefix?"', prefix: '":"", rxc->prefix?rxc->prefix:"");
         free(rs_str);
         free(attr_str);
     }
     for(r=tree->neg_rx_lst;r!=NULL;r=r->next) {
         rxc=r->data;
-        log_msg(log_level, "%-*s  '!%s %s' (%s:%d: '%s')", depth+2, "\u2502", rxc->rx, rs_str = get_restriction_string(rxc->restriction), rxc->config_filename, rxc->config_linenumber, rxc->config_line);
+        log_msg(log_level, "%-*s  '!%s %s' (%s:%d: '%s%s%s')", depth+2, "\u2502", rxc->rx, rs_str = get_restriction_string(rxc->restriction), rxc->config_filename, rxc->config_linenumber, rxc->config_line, rxc->prefix?"', prefix: '":"", rxc->prefix?rxc->prefix:"");
         free(rs_str);
     }
 
@@ -73,7 +86,6 @@ void log_tree(LOG_LEVEL log_level, seltree* tree, int depth) {
 static char* strrxtok(char* rx)
 {
   char*p=NULL;
-  char*t=NULL;
   size_t i=0;
 
   /* The following code assumes that the first character is a slash */
@@ -97,10 +109,9 @@ static char* strrxtok(char* rx)
 	i=strlen(p);
 	break;
       case '\\':
-	t=checked_strdup(p);
-	strcpy(p+i,t+i+1);
-	free(t);
-	t=NULL;
+        for (int j = i; p[j]; j++) {
+            p[j] = p[j+1];
+        }
 	break;
       default:
 	break;
@@ -176,15 +187,18 @@ int compare_node_by_path(const void *n1, const void *n2)
 
 seltree* get_seltree_node(seltree* tree,char* path)
 {
+  LOG_LEVEL log_level = LOG_LEVEL_TRACE;
   seltree* node=NULL;
   list* r=NULL;
   char* tmp=NULL;
 
   if(tree==NULL){
+    log_msg(log_level, "get_seltree_node(): return NULL (tree == NULL)");
     return NULL;
   }
 
   if(strncmp(path,tree->path,strlen(path)+1)==0){
+    log_msg(log_level, "get_seltree_node(): return %p (path: %s)", tree, tree->path);
     return tree;
   }
   else{
@@ -195,12 +209,14 @@ seltree* get_seltree_node(seltree* tree,char* path)
 	if(node!=NULL){
 	  /* Don't leak memory */
 	  free(tmp);
+	  log_msg(log_level, "get_seltree_node(): return '%s' (%p, path: '%s')", node->path, node, path);
 	  return node;
 	}
       }
     }
     free(tmp);
   }
+  log_msg(log_level, "get_seltree_node(): return NULL (path: '%s')", path);
   return NULL;
 }
 
@@ -250,7 +266,7 @@ seltree* new_seltree_node(
   }else {
     node->parent=NULL;
   }
-  log_msg(LOG_LEVEL_DEBUG, "new node '%s' (%p, parent: %p)", node->path, node, node->parent);
+  log_msg(LOG_LEVEL_TRACE, "new node '%s' (%p, parent: %p)", node->path, node, node->parent);
   return node;
 }
 
@@ -260,7 +276,7 @@ seltree *init_tree() {
     return node;
 }
 
-rx_rule * add_rx_to_tree(char * rx, RESTRICTION_TYPE restriction, int rule_type, seltree *tree, const char **rule_error, int *rule_erroffset) {
+rx_rule * add_rx_to_tree(char * rx, RESTRICTION_TYPE restriction, int rule_type, seltree *tree, int linenumber, char* filename, char* linebuf) {
     rx_rule* r = NULL;
     seltree *curnode = NULL;
     char *rxtok = NULL;
@@ -275,17 +291,36 @@ rx_rule * add_rx_to_tree(char * rx, RESTRICTION_TYPE restriction, int rule_type,
     r->config_linenumber = -1;
     r->attr = 0;
 
-    if((r->crx=pcre_compile(r->rx, PCRE_ANCHORED|PCRE_UTF8, rule_error, rule_erroffset, NULL)) == NULL) {
+    int pcre2_errorcode;
+    PCRE2_SIZE pcre2_erroffset;
+
+    if((r->crx=pcre2_compile((PCRE2_SPTR) r->rx, PCRE2_ZERO_TERMINATED, PCRE2_UTF|PCRE2_ANCHORED, &pcre2_errorcode, &pcre2_erroffset, NULL)) == NULL) {
+        PCRE2_UCHAR pcre2_error[128];
+        pcre2_get_error_message(pcre2_errorcode, pcre2_error, 128);
+        log_msg(LOG_LEVEL_ERROR, "%s:%d:%i: error in rule '%s': %s (line: '%s')", filename, linenumber, pcre2_erroffset, rx, pcre2_error, linebuf);
         free(r);
         return NULL;
     } else {
+        r->md = pcre2_match_data_create_from_pattern(r->crx, NULL);
+        if (r->md == NULL) {
+            log_msg(LOG_LEVEL_ERROR, "pcre2_match_data_create_from_pattern: failed to allocate memory");
+            exit(MEMORY_ALLOCATION_FAILURE);
+        }
+        int pcre2_jit = pcre2_jit_compile(r->crx, PCRE2_JIT_PARTIAL_SOFT);
+        if (pcre2_jit < 0) {
+            PCRE2_UCHAR pcre2_error[128];
+            pcre2_get_error_message(pcre2_jit, pcre2_error, 128);
+            log_msg(LOG_LEVEL_NOTICE, "JIT compilation for regex '%s' failed: %s (fall back to interpreted matching)", r->rx, pcre2_error);
+        } else {
+            log_msg(LOG_LEVEL_DEBUG, "JIT compilation for regex '%s' successful", r->rx);
+        }
+
         rxtok=strrxtok(r->rx);
         curnode=get_seltree_node(tree,rxtok);
 
         for(size_t i=1;i < strlen(rxtok); ++i){
             if (rxtok[i] == '/' && rxtok[i-1] == '/') {
-                *rule_error = "invalid double slash" ;
-                *rule_erroffset = i;
+                log_msg(LOG_LEVEL_ERROR, "%s:%d:1: error in rule '%s': invalid double slash (line: '%s')", filename, linenumber, rx, linebuf);
                 free(r);
                 return NULL;
             }
@@ -316,21 +351,20 @@ rx_rule * add_rx_to_tree(char * rx, RESTRICTION_TYPE restriction, int rule_type,
 }
 
 #define LOG_MATCH(log_level, border, format, ...) \
-    log_msg(log_level, "%s %*c'%s' " #format " of %s (%s:%d: '%s')", border, depth+2, ' ', text, __VA_ARGS__, get_rule_type_long_string(rule_type), rx->config_filename, rx->config_linenumber, rx->config_line);
+    log_msg(log_level, "%s %*c'%s' " #format " of %s (%s:%d: '%s%s%s')", border, depth+2, ' ', text, __VA_ARGS__, get_rule_type_long_string(rule_type), rx->config_filename, rx->config_linenumber, rx->config_line, rx->prefix?"', prefix: '":"", rx->prefix?rx->prefix:"");
 
 static int check_list_for_match(list* rxrlist,char* text, rx_rule* *rule, RESTRICTION_TYPE file_type, int rule_type, int depth, bool unrestricted_only)
 {
   list* r=NULL;
   int retval=NO_RULE_MATCH;
   int pcre_retval;
-  pcre_extra *pcre_extra = NULL;
   char *rs_str = NULL;
   for(r=rxrlist;r;r=r->next){
       rx_rule *rx = (rx_rule*)r->data;
 
       if (!(unrestricted_only && rx->restriction)) {
 
-      pcre_retval=pcre_exec((pcre*)rx->crx, pcre_extra, text, strlen(text), 0, PCRE_PARTIAL_SOFT, NULL, 0);
+      pcre_retval = pcre2_match(rx->crx, (PCRE2_SPTR) text, PCRE2_ZERO_TERMINATED, 0, PCRE2_PARTIAL_SOFT, rx->md, NULL);
       if (pcre_retval >= 0) {
           if (!rx->restriction || file_type&rx->restriction) {
                   *rule = rx;
@@ -343,7 +377,7 @@ static int check_list_for_match(list* rxrlist,char* text, rx_rule* *rule, RESTRI
               free(rs_str);
               retval=PARTIAL_RULE_MATCH;
           }
-      } else if (pcre_retval == PCRE_ERROR_PARTIAL) {
+      } else if (pcre_retval == PCRE2_ERROR_PARTIAL) {
           LOG_MATCH(LOG_LEVEL_RULE, "\u2502", partially matches regex '%s', rx->rx)
           retval=PARTIAL_RULE_MATCH;
       } else {
@@ -385,8 +419,8 @@ static int check_node_for_match(seltree *node, char *text, RESTRICTION_TYPE file
 
   /* if this call is not recursive we check the equals list and we set top *
    * and retval so we know following calls are recursive */
-  if(!(retval&16)){
-      retval|=16;
+  if(!(retval&RECURSED_CALL)){
+      retval|=RECURSED_CALL;
 
       if (node->equ_rx_lst) {
           log_msg(LOG_LEVEL_RULE, "\u2502 %*cnode: '%s': check equal list", depth, ' ', node->path);
@@ -394,14 +428,11 @@ static int check_node_for_match(seltree *node, char *text, RESTRICTION_TYPE file
               case RESTRICTED_RULE_MATCH:
               case RULE_MATCH: {
                           log_msg(LOG_LEVEL_RULE, "\u2502 %*cequal match for '%s' (node: '%s')", depth, ' ', text, node->path);
-                          retval|=2|4;
+                          retval|=EQUAL_RULE_MATCH|DEEP_EQUAL_MATCH;
                           break;
                       }
               case PARTIAL_RULE_MATCH: {
-                           if(file_type&FT_DIR && get_seltree_node(node,text)==NULL) {
-                               seltree *new_node = new_seltree_node(node,text,0,NULL);
-                               log_msg(LOG_LEVEL_DEBUG, "added new node '%s' (%p) for '%s' (reason: partial equal match for directory)", new_node->path, new_node, text);
-                           }
+                           retval|=PARTIAL_RULE_MATCH;
                            break;
                        }
           }
@@ -414,22 +445,19 @@ static int check_node_for_match(seltree *node, char *text, RESTRICTION_TYPE file
   /* We'll use retval to pass information on whether to recurse
    * the dir or not */
 
-  /* If 4 and 8 are not set, we will check for matches */
-  if(!(retval&(4|8))){
+  /* If we have no deep matches, we will check for matches */
+  if(!(retval&(DEEP_EQUAL_MATCH|DEEP_SELECTIVE_MATCH))){
       if (node->sel_rx_lst) {
           log_msg(LOG_LEVEL_RULE, "\u2502 %*cnode: '%s': check selective list", depth, ' ', node->path);
           switch (check_list_for_match(node->sel_rx_lst, text, rule, file_type, AIDE_SELECTIVE_RULE, depth, false)) {
               case RESTRICTED_RULE_MATCH:
               case RULE_MATCH: {
                           log_msg(LOG_LEVEL_RULE, "\u2502 %*cselective match for '%s' (node: '%s')", depth, ' ', text, node->path);
-                          retval|=1|8;
+                          retval|=SELECtIVE_RULE_MATCH|DEEP_SELECTIVE_MATCH;
                           break;
                       }
               case PARTIAL_RULE_MATCH: {
-                           if(file_type&FT_DIR && get_seltree_node(node,text)==NULL) {
-                               seltree *new_node = new_seltree_node(node,text,0,NULL);
-                               log_msg(LOG_LEVEL_DEBUG, "added new node '%s', (%p) for '%s' (reason: partial selective match for directory)", new_node->path, new_node, text);
-                           }
+                           retval|=PARTIAL_RULE_MATCH;
                            break;
                        }
           }
@@ -441,11 +469,11 @@ static int check_node_for_match(seltree *node, char *text, RESTRICTION_TYPE file
   }
 
   /* Now let's check the ancestors */
-  retval=check_node_for_match(node->parent, text, file_type, retval&~32, rule, depth+2);
+  retval=check_node_for_match(node->parent, text, file_type, retval&~TOP_LEVEL_CALL, rule, depth+2);
 
   /* Negative regexps are the strongest so they are checked last */
   /* If this file is to be added */
-  if(retval&(1|2)){
+  if(retval&(SELECtIVE_RULE_MATCH|EQUAL_RULE_MATCH)){
       if (node->neg_rx_lst) {
           log_msg(LOG_LEVEL_RULE, "\u2502 %*cnode: '%s': check negative list (reason: previous positive match)", depth, ' ', node->path);
 
@@ -461,27 +489,23 @@ static int check_node_for_match(seltree *node, char *text, RESTRICTION_TYPE file
                   log_msg(LOG_LEVEL_RULE, "\u2502 %*ccheck parent directory '%s' (unrestricted rules only)", depth+2, ' ', parentname);
                   if (check_list_for_match(node->neg_rx_lst, parentname, rule, FT_DIR, AIDE_NEGATIVE_RULE, depth+4, true) == RULE_MATCH) {
                       log_msg(LOG_LEVEL_RULE, "\u2502 %*cnegative match for parent directory '%s'", depth, ' ', parentname);
-                      retval=0;;
+                      retval=NEGATIVE_RULE_MATCH;
                       break;
                   }
               }
           } while (strcmp(parentname,node->path) > 0);
           free(parentname);
 
-          if (retval) {
+          if (retval != NEGATIVE_RULE_MATCH) {
           log_msg(LOG_LEVEL_RULE, "\u2502 %*ccheck file '%s'", depth+2, ' ', text);
           switch (check_list_for_match(node->neg_rx_lst, text, rule, file_type, AIDE_NEGATIVE_RULE, depth+2, false)) {
               case RESTRICTED_RULE_MATCH: {
-                  if(file_type&FT_DIR && get_seltree_node(node,text)==NULL) {
-                      seltree *new_node = new_seltree_node(node,text,0,NULL);
-                      log_msg(LOG_LEVEL_DEBUG, "added new node '%s' (%p) for '%s' (reason: restricted negative match for directory)", new_node->path, new_node, text);
-                  }
-
+                  retval=PARTIAL_RULE_MATCH;
+                  break;
               }
-              // fall through
               case RULE_MATCH: {
                   log_msg(LOG_LEVEL_RULE, "\u2502 %*cnegative match for '%s' (node: '%s')", depth, ' ', text, node->path);
-                  retval=0;
+                  retval=NEGATIVE_RULE_MATCH;
                   break;
               }
           }
@@ -497,13 +521,17 @@ static int check_node_for_match(seltree *node, char *text, RESTRICTION_TYPE file
 
   } else {
     log_msg(LOG_LEVEL_DEBUG, "\u2502 %*cskip node '%s' (reason: no regex rules)", depth, ' ', node->path);
-    retval = check_node_for_match(node->parent, text, file_type, (retval|16)&~32, rule, depth);
+    retval = check_node_for_match(node->parent, text, file_type, (retval|RECURSED_CALL)&~TOP_LEVEL_CALL, rule, depth);
   }
 
   /* Now we discard the info whether a match was made or not *
    * and just return 0,1 or 2 */
-  if(!(retval&32)){
-      retval&=3;
+  if(!(retval&TOP_LEVEL_CALL)){
+      if (retval&(EQUAL_RULE_MATCH|SELECtIVE_RULE_MATCH)) {
+        retval&=(EQUAL_RULE_MATCH|SELECtIVE_RULE_MATCH);
+      } else {
+        retval&=PARTIAL_RULE_MATCH;
+      }
   }
   return retval;
 }
@@ -523,37 +551,36 @@ int check_seltree(seltree *tree, char *filename, RESTRICTION_TYPE file_type, rx_
   if(tmp!=parentname){
     *tmp='\0';
   }else {
-
-    if(parentname[1]!='\0'){
       /* we are in the root dir */
       parentname[1]='\0';
-    }
   }
 
+  log_msg(LOG_LEVEL_TRACE, "\u2502 search for parent node '%s' (tree: '%s' (%p))", parentname, tree->path, tree);
   pnode=get_seltree_node(tree,parentname);
   if (pnode == NULL) {
-    retval |= 16;
+    retval |= RECURSED_CALL;
   }
 
   } while (pnode == NULL);
 
-  log_msg(LOG_LEVEL_DEBUG, "got parent node '%s' (%p) for parentname '%s'", pnode->path, pnode, parentname);
+  log_msg(LOG_LEVEL_DEBUG, "\u2502 got parent node '%s' (%p) for parentname '%s'", pnode->path, pnode, parentname);
 
   free(parentname);
 
-  retval = check_node_for_match(pnode, filename, file_type, retval|32 ,rule, 0);
-
-  if (retval) {
-    char *str;
-    log_msg(LOG_LEVEL_RULE, "\u2534 ADD '%s' to the tree (attr: '%s')", filename, str = diff_attributes(0, (*rule)->attr));
-    free(str);
+  retval = check_node_for_match(pnode, filename, file_type, retval|TOP_LEVEL_CALL ,rule, 0);
 
+  if (retval&(SELECtIVE_RULE_MATCH|EQUAL_RULE_MATCH)) {
     if(get_seltree_node(tree,filename)==NULL) {
         seltree *new_node = new_seltree_node(tree,filename,0,NULL);
-        log_msg(LOG_LEVEL_DEBUG, "added new node '%s', (%p) for '%s' (reason: full match)", new_node->path, new_node, filename);
+        log_msg(LOG_LEVEL_DEBUG, "\u2502 added new node '%s', (%p) for '%s' (reason: full match)", new_node->path, new_node, filename);
     }
+
+    char *str;
+    log_msg(LOG_LEVEL_RULE, "\u2534 ADD '%s' (attr: '%s')", filename, str = diff_attributes(0, (*rule)->attr));
+    free(str);
   } else {
-    log_msg(LOG_LEVEL_RULE, "\u2534 do NOT add '%s' to the tree", filename);
+    log_msg(LOG_LEVEL_RULE, "\u2534 do NOT add '%s'", filename);
   }
+  log_msg(LOG_LEVEL_TRACE, "check_seltree: return %d for '%s'", retval, filename);
   return retval;
 }
