@@ -1,7 +1,7 @@
 /*
  * AIDE (Advanced Intrusion Detection Environment)
  *
- * Copyright (C) 1999-2002, 2004-2006, 2010-2011, 2013, 2016, 2019-2021
+ * Copyright (C) 1999-2002, 2004-2006, 2010-2011, 2013, 2016, 2019-2022
  *               Rami Lehti, Pablo Virolainen, Mike Markley,
  *               Richard van den Berg, Hannes von Haugwitz
  *
@@ -20,12 +20,17 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
+#include "config.h"
 #include <stdlib.h>
-#include <unistd.h>
+#include <stdio.h>
 #include <string.h>
 #include <sys/stat.h>
 #include <ctype.h>
 #include <syslog.h>
+#include <stdbool.h>
+#include <strings.h>
+#include <unistd.h>
+#include "url.h"
 /*for locale support*/
 #include "locale-aide.h"
 /*for locale support*/
@@ -38,6 +43,7 @@
 #include "log.h"
 #include "db_config.h"
 #include "util.h"
+#include "errorcodes.h"
 
 #define URL_UNSAFE " <>\"#%{}|\\^~[]`@:\033'"
 #define ISPRINT(c) (isascii(c) && isprint(c))
@@ -50,7 +56,7 @@ void* checked_malloc(size_t size) {
     void * p = malloc(size);
     if (p == NULL) {
         log_msg(LOG_LEVEL_ERROR, "malloc: failed to allocate %d bytes of memory", size);
-        exit(EXIT_FAILURE);
+        exit(MEMORY_ALLOCATION_FAILURE);
     }
     return p;
 }
@@ -58,7 +64,7 @@ void* checked_calloc(size_t nmemb, size_t size) {
     void * p = calloc(nmemb, size);
     if (p == NULL) {
         log_msg(LOG_LEVEL_ERROR, "calloc: failed to allocate %d bytes of memory", size);
-        exit(EXIT_FAILURE);
+        exit(MEMORY_ALLOCATION_FAILURE);
     }
     return p;
 }
@@ -66,7 +72,7 @@ void* checked_strdup(const char *s) {
     void * p = strdup(s);
     if (p == NULL) {
         log_msg(LOG_LEVEL_ERROR, "strdup: failed to allocate memory");
-        exit(EXIT_FAILURE);
+        exit(MEMORY_ALLOCATION_FAILURE);
     }
     return p;
 }
@@ -74,7 +80,7 @@ void* checked_strndup(const char *s, size_t size) {
     void * p = strndup(s, size);
     if (p == NULL) {
         log_msg(LOG_LEVEL_ERROR, "strdup: failed to allocate memory");
-        exit(EXIT_FAILURE);
+        exit(MEMORY_ALLOCATION_FAILURE);
     }
     return p;
 }
@@ -82,7 +88,7 @@ void* checked_realloc(void *ptr, size_t size) {
     void * p = realloc(ptr,size);
     if (p == NULL) {
         log_msg(LOG_LEVEL_ERROR, "realloc: failed to allocate memory");
-        exit(EXIT_FAILURE);
+        exit(MEMORY_ALLOCATION_FAILURE);
     }
     return p;
 }
@@ -97,7 +103,7 @@ int cmpurl(url_t* u1,url_t* u2)
   }
 
   return RETOK;
-};
+}
 
 /* Returns 1 if the string contains unsafe characters, 0 otherwise.  */
 int contains_unsafe (const char *s)
@@ -264,17 +270,16 @@ char* perm_to_char(mode_t perm)
 char *expand_tilde(char *path) {
     char *homedir = NULL;
     char *full = NULL;
-    size_t path_len, homedir_len, full_len;
 
     if (path != NULL) {
         if (path[0] == '~') {
             if((homedir=getenv("HOME")) != NULL) {
+                size_t path_len, homedir_len, full_len;
                 path_len = strlen(path+sizeof(char));
                 homedir_len = strlen(homedir);
-                full_len = homedir_len+path_len;
-                full = checked_malloc(sizeof(char) * (full_len+1));
-                strcpy(full, homedir);
-                strcat(full+homedir_len, path+sizeof(char));
+                full_len = (homedir_len+path_len+1) * sizeof(char);
+                full = checked_malloc(full_len);
+                snprintf(full, full_len, "%s%s", homedir, path+sizeof(char));
                 log_msg(LOG_LEVEL_DEBUG, "expanded '~' in '%s' to '%s'", path, full);
                 free(path);
                 /* Don't free(homedir); because it is not safe on some platforms */
@@ -289,6 +294,21 @@ char *expand_tilde(char *path) {
     return path;
 }
 
+char* pipe2string(int fd) {
+    int nbytes;
+    char buffer[1024];
+
+    char* str = NULL;
+    while ((nbytes = read(fd, buffer, sizeof(buffer))) > 0) {
+        int str_len = str?strlen(str):0U;
+        int len = str_len+nbytes;
+        str = checked_realloc(str, (len+1)*sizeof(char));
+        strncpy(str+str_len, buffer, nbytes);
+        str[len] = '\0';
+    }
+    return str;
+}
+
 /* Like strstr but only do search for maximum of n chars.
    haystack does not have to be NULL terminated
    needle has to be NULL terminated. NULL in needle is not used in compare.
@@ -338,6 +358,7 @@ size_t strnlen(const char *s, size_t maxlen)
 }
 #endif
 
+#ifdef HAVE_SYSLOG
 /* Lookup syslog facilities by name */
 int syslog_facility_lookup(char *s)
 {
@@ -419,3 +440,4 @@ int syslog_facility_lookup(char *s)
 	log_msg(LOG_LEVEL_WARNING, "Syslog facility \"%s\" is unknown, using default",s);
 	return(AIDE_SYSLOG_FACILITY);
 }
+#endif
